%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%% Written by Zheng Fang -- December 31 2020 %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Data source (last accessed: January 2, 2021): 
% http://qed.econ.queensu.ca/jae/datasets/beare001/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Load Data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
warning off
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',1231);
RandStream.setGlobalStream(stream);
% Select shape to be tested
Shape = 1;
% Load data
DataSet = readmatrix('option_prices.txt');
Option = 1; % 0 for call option and 1 for put option
if Option == 0
    Data = DataSet(DataSet(:,1) == 2008 & DataSet(:,2) == 10 & ...
        DataSet(:,3) == 22   & DataSet(:,10) == Option,:);
elseif Option == 1
    Data = DataSet(DataSet(:,1) == 2011 & DataSet(:,2) == 10 & ...
        DataSet(:,3) == 19   & DataSet(:,10) == Option,:);
else
    error('Option must be either 0 or 1')
end
Z = Data(:,5);                      % Strike price
Y = Data(:,7);                      % Closing price of option
N = length(Z);                      % Extract the sample size
Z = 2*(Z-min(Z))/(max(Z)-min(Z))-1; % Normalize the range

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Parameters to Choose %%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Boot = 1000;                    % Number of bootstrap repetitions
Zkn = 2;                        % Number of interior knots for B-splines
Ord = 3;                        % Order of B-splines: 4,5
Zmin = -0.9;                    % Left endpoint for the restriction
Zmax = 0.9;                     % Right endpoint for the restriction
GridStep = 0.01;                % Step size for discretization             
Bandwidth = 'L3';               % Select the tuning parameter
if strcmpi(Bandwidth,'C1')
    gamma = 0.1;                     
elseif strcmpi(Bandwidth,'C2')
    gamma = 0.05;
elseif strcmpi(Bandwidth,'C3')
    gamma = 0.01;
elseif strcmpi(Bandwidth,'L1')
    gamma = 0.1/log(N);
elseif strcmpi(Bandwidth,'L2')
    gamma = 0.05/log(N);
elseif strcmpi(Bandwidth,'L3')
    gamma = 0.01/log(N);
elseif strcmpi(Bandwidth,'P1')
    gamma = 1/N^0.5;
elseif strcmpi(Bandwidth,'P2')
    gamma = 1/N^0.75;
elseif strcmpi(Bandwidth,'P3')
    gamma = 1/N;
else
    error('Try a different Bandwidth')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Parameters Generated from the Choices %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
kn = Zkn + Ord;                 % The sieve dimension for Z
Xin = sqrt(kn);                 % Norm bound of B-splines
Grid = (Zmin:GridStep:Zmax)';   % The vector of grid points
T = length(Grid);               % Number of the grid points 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create A and b for the constraint select in format required by OSQP
A1 = eye(T-1,T) - [zeros(T-1,1) eye(T-1,T-1)];  
if Shape  == 1                                % Test monotonicity
    if Option == 0
        A = -A1;                              % Decreasing for call
    else
        A = A1;                               % Increasing for put
    end        
elseif Shape == 2                             % Test convexity 
    A2 = eye(T-2,T-1) - [zeros(T-2,1) eye(T-2,T-2)];
    A = A2*(-A1);                         
    clear  A1 A2
elseif Shape == 3                             % Test both
    A2 = eye(T-2,T-1) - [zeros(T-2,1) eye(T-2,T-2)];
    if Option == 0
        A = [-A1; -A2*A1];
    else
        A = [A1; -A2*A1];
    end
    clear  A1 A2
end
lb = -Inf(size(A,1),1);
ub = zeros(size(A,1),1);
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Implement the Test %%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct basis functions. First, obtain the knot sequences for the
% B-splines based on equiquantile spacing interior knots
Zknots = quantile(Z,0:1/(Zkn+1):1);
Zknots = augknt(Zknots,Ord);                  % The augmented knots
% Evaluate the resulting kn-dimensional B-splines at the grid points
Sieve = spcol(Zknots, Ord, Grid);             % Size: T*kn

% Construct the matrix of Splines evaluated at the data
[ZTemp, index] = sort(Z);       % Sorting required by spcol
Hn = spcol(Zknots, Ord, ZTemp); % Size: N*kn
Hn = sortrows([index Hn]);      % Arrange rows in Hn from i=1,...,n
Hn = Hn(:,2:end);               % Drop the index column

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% First, obtain the (unconstrained) series 2SLS estimators
Hatbeta = Hn\Y;           % Coefficient
Hattheta = Sieve*Hatbeta; % Unconstrained estimator
HatU = Y - Hn*Hatbeta;    % Residuals

% In turn, obtain the constrained estimator through projection
Prob = osqp;              % Initialize the solver
Prob.setup(eye(T),-Hattheta,A,lb,ub,'eps_abs',1e-05,'eps_rel', 1e-05,...
                          'rho',1/T,'verbose',0);  % Specify the problem
Results = Prob.solve();   % Just solve it!
Projtheta = Results.x;    % Get the projection as the minimzier

% Finally, compute the test statistic via numerical integration based on 
% the trapezoidal integration
Stat = (sqrt(N)/Xin)*trapz(GridStep,(Hattheta - Projtheta).^2)^0.5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bootstrap Gn
BootW = randn(N,Boot);                        % Bootstrap weights
BGn = (sqrt(N)/Xin)*Sieve*(Hn\(BootW.*HatU)); % Score bootstrap: T*Boot

% Next, obtain the data-driven tuning parameter
Hattau = quantile(trapz(GridStep,BGn.^2).^0.5,1-gamma); % Quantile
Hatkappa = sqrt(N)/(log(N)*Xin*Hattau);       % $\hat\kappa_n$

% Now, construct the bootstrap statistics
BootSum = BGn+Hatkappa*Projtheta;   % Bootstrap + drift: T*Boot
BootProj = zeros(T,Boot);           % Projection of BootSum
parfor i=1:Boot
    Prob = osqp;                    % Initialize the solver
    Prob.setup(eye(T),-BootSum(:,i),A,lb,ub,'eps_abs',1e-05,...
        'eps_rel',1e-05,'rho',1/T,'verbose',0); % Specify the problem
    BootResults = Prob.solve();     % Just solve it!
    BootProj(:,i) = BootResults.x;  % Get the projection as the minimizer
end
BStats = trapz(GridStep,(BootSum-BootProj).^2).^0.5; % Size: Boot*1

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Record p-Value %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvalue = mean(BStats >= Stat);

disp(['The p-value is ' num2str(pvalue)])
 
% Close parallel computing 
% delete (gcp);