%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Written by Zheng Fang and Juwon Seo -- March 2020 %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements our shape tests of the regression function in 
% univariate NP models, based on B-splines. The "Shape" indicator allows 
% one to choose different shapes to test (monotonicity, convexity, or both). 
% We solve the quadratic problems by the solver OSQP (https://osqp.org/).

% One needs to i) properly load the data, and ii) choose the some parameters
% in the "Parameters to Choose" block. The output is p-value of the test.

% A (very) brief introduction to B-splines may be found in Appendix E.1 in
% Chen & Christensen (2018, QE). One may also want to consult de Boor's "A
% Practical Guide to Splines" (2001), in particular Theorem 44 on p.97 and
% Figure 50 on p.100, to better understand the commmands augknt and spcol.

% The MATLAB Engine API for Python allows one to call MATLAB as a 
% computational engine from Python; see instructutions at
% https://www.mathworks.com/products/matlab/matlab-and-python.html.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Load Data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% You could import data interactively by "Importa Data" under "Home". Or,
load('Sample1.mat');                     % For mat file
% Data = readmatrix('Sample1.txt');      % For txt file
% Data = readmatrix('Sample1.xlsx');     % For xlsx file

% Below, we assume variables Y and Z are in the Workspace (by renaming 
% variables if necessary). 
N = length(Y);                           % Obtain the sample size 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Parameters to Choose %%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                   % The significance level
Boot = 200;                     % Number of bootstrap repetitions
Shape = 1;                      % 1 for monotonicity, 2 convexity, 3 both
Zkn = 3;                        % Number of interior knots for B-splines
Ord = 4;                        % Order of B-splines: 4,5
Zmin = -0.9;                    % Left endpoint for the restriction
Zmax = 0.9;                     % Right endpoint for the restriction
GridStep = 0.05;                % Step size for discretization             
Bandwidth = 'L3';               % Select the tuning parameter
if strcmpi(Bandwidth,'C1')
    gamma = 0.1;                     
elseif strcmpi(Bandwidth,'C2')
    gamma = 0.05;
elseif strcmpi(Bandwidth,'C3')
    gamma = 0.01;
elseif strcmpi(Bandwidth,'L1')
    gamma = 0.1/log(N);
elseif strcmpi(Bandwidth,'L2')
    gamma = 0.05/log(N);
elseif strcmpi(Bandwidth,'L3')
    gamma = 0.01/log(N);
elseif strcmpi(Bandwidth,'P1')
    gamma = 1/N^0.5;
elseif strcmpi(Bandwidth,'P2')
    gamma = 1/N^0.75;
elseif strcmpi(Bandwidth,'P3')
    gamma = 1/N;
else
    error('Try a different Bandwidth')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Parameters Generated from the Choices %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
kn = Zkn+Ord;                   % The sieve dimension for Z
Xin = sqrt(kn);                 % Norm bound of B-splines
Grid = (Zmin:GridStep:Zmax)';   % The vector of grid points
T = length(Grid);               % Number of the grid points 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create A and b for the constraint select in format required by OSQP
A1 = eye(T-1,T) - [zeros(T-1,1) eye(T-1,T-1)];  
if Shape  == 1                                % Test monotonicity
    A = A1;
elseif Shape == 2                             % Test convexity
    A2 = eye(T-2,T-1) - [zeros(T-2,1) eye(T-2,T-2)];
    A = A2*(-A1);                             % Be careful about the sign!!
    clear  A1 A2
elseif Shape == 3                             % Test both
    A2 = eye(T-2,T-1) - [zeros(T-2,1) eye(T-2,T-2)];
    A = [A1; -A2*A1];
    clear  A1 A2
end
lb = -Inf(size(A,1),1);              % Bounds of constraints for the solver
ub = zeros(size(A,1),1);
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Implement the Test %%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct basis functions. First, obtain the knot sequences for the
% B-splines based on equiquantile spacing interior knots
Zknots = quantile(Z,0:1/(Zkn+1):1);
Zknots = augknt(Zknots,Ord);                  % The augmented knots
% Evaluate the resulting kn-dimensional B-splines at the grid points
Sieve = spcol(Zknots, Ord, Grid);             % Size: T*kn

% Construct the matrix of Splines evaluated at the data
[ZTemp, index] = sort(Z);       % Sorting required by spcol
Hn = spcol(Zknots, Ord, ZTemp); % Size: N*kn
Hn = sortrows([index Hn]);      % Arrange rows in Hn from i=1,...,n
Hn = Hn(:,2:end);               % Drop the index column

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% First, obtain the (unconstrained) series 2SLS estimators
Hatbeta = Hn\Y;           % Coefficient
Hattheta = Sieve*Hatbeta; % Unconstrained estimator
HatU = Y - Hn*Hatbeta;    % Residuals

% In turn, obtain the constrained estimator through projection
Prob = osqp;              % Initialize the solver
Prob.setup(eye(T),-Hattheta,A,lb,ub,'eps_abs',1e-04,'eps_rel', 1e-04,...
                          'rho',1/T,'verbose',0);  % Specify the problem
Results = Prob.solve();   % Just solve it!
Projtheta = Results.x;    % Get the projection as the minimzier

% Finally, compute the test statistic via numerical integration based on 
% the trapezoidal integration
Stat = (sqrt(N)/Xin)*trapz(GridStep,(Hattheta - Projtheta).^2)^0.5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bootstrap Gn
BootW = randn(N,Boot);                        % Bootstrap weights
BGn = (sqrt(N)/Xin)*Sieve*(Hn\(BootW.*HatU)); % Score bootstrap: T*Boot

% Next, obtain the data-driven tuning parameter
Hattau = quantile(trapz(GridStep,BGn.^2).^0.5,1-gamma); % Quantile
Hatkappa = sqrt(N)/(log(N)*Xin*Hattau);       % $\hat\kappa_n$

% Now, construct the bootstrap statistics
BootSum = BGn+Hatkappa*Projtheta;   % Bootstrap + drift: T*Boot
BootProj = zeros(T,Boot);           % Projection of BootSum
parfor i=1:Boot
    Prob = osqp;                    % Initialize the solver
    Prob.setup(eye(T),-BootSum(:,i),A,lb,ub,'eps_abs',1e-04,...
        'eps_rel',1e-04,'rho',1/T,'verbose',0); % Specify the problem
    BootResults = Prob.solve();     % Just solve it!
    BootProj(:,i) = BootResults.x;  % Get the projection as the minimizer
end
BStats = trapz(GridStep,(BootSum-BootProj).^2).^0.5; % Size: Boot*1

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Record p-Value %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvalue = mean(BStats >= Stat);

disp(['The p-value is ' num2str(pvalue)])
 
% Close parallel computing 
delete (gcp);