%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Written by Zheng Fang and Juwon Seo -- April 2020 %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements our shape tests of the regression function in NP 
% models with multiple regressors, based on B-splines. The "Shape" 
% indicator allows one to choose different shapes to test (monotonicity, 
% convexity, or both). We solve the quadratic problems by the solver OSQP 
% (https://osqp.org/).

% One needs to i) properly load the data, and ii) choose the some parameters
% in the "Parameters to Choose" block. The output is p-value of the test.

% A (very) brief introduction to B-splines may be found in Appendix E.1 in
% Chen & Christensen (2018, QE). One may also want to consult de Boor's "A
% Practical Guide to Splines" (2001), in particular Theorem 44 on p.97 and
% Figure 50 on p.100, to better understand the commmands augknt and spcol.

% The MATLAB Engine API for Python allows one to call MATLAB as a 
% computational engine from Python; see instructutions at
% https://www.mathworks.com/products/matlab/matlab-and-python.html.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Load Data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% You could import data interactively by "Importa Data" under "Home". Or,
load('Sample2.mat');                     % For mat file
% Data = readmatrix('Sample2.txt');      % For txt file
% Data = readmatrix('Sample2.xlsx');     % For xlsx file

% Below, we assume variables Y and Z are in the Workspace (by renaming 
% variables if necessary). 
N = length(Y);                           % Obtain the sample size 
Zdim = size(Z,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Parameters to Choose %%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                  % The significance level
Boot = 200;                    % Number of bootstrap repetitions
Shape = 3;                     % 1 for monotonicity, 2 concavity, 3 both
Zkn = 1;                       % Number of interior knots for B-splines
Ord = 3;                       % Order of B-splines 
% Below we assume the marginal interval is the same for the constraint, but
% one can adjust this as needed.
Zmin = 0.1;                    % Left endpoint for the restriction
Zmax = 0.9;                    % Right endpoint for the restriction
if Shape == 1
    GridStep = 0.05;           % Step size for discretization
else
    GridStep = 0.1;            % Step size for discretization
end
Bandwidth = 'L3';              % Select the tuning parameter
if strcmpi(Bandwidth,'C1')
    gamma = 0.1;                     
elseif strcmpi(Bandwidth,'C2')
    gamma = 0.05;
elseif strcmpi(Bandwidth,'C3')
    gamma = 0.01;
elseif strcmpi(Bandwidth,'L1')
    gamma = 0.1/log(N);
elseif strcmpi(Bandwidth,'L2')
    gamma = 0.05/log(N);
elseif strcmpi(Bandwidth,'L3')
    gamma = 0.01/log(N);
elseif strcmpi(Bandwidth,'P1')
    gamma = 1/N^0.5;
elseif strcmpi(Bandwidth,'P2')
    gamma = 1/N^0.75;
elseif strcmpi(Bandwidth,'P3')
    gamma = 1/N;
else
    error('Try a different Bandwidth')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Parameters Generated from the Choices %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
kn = Zkn+Ord;                  % Sieve dimension for Z
Xin = sqrt(kn^Zdim);           % Norm bound of the tensor product B-splines
Grid = (Zmin:GridStep:Zmax)';  % The vector of grid points
T0 = length(Grid);             % Number of grid points along each dimension
T = T0^Zdim;                   % Total number of grid points

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Next, configurations for projection depending on the shape selected.
% Start with the monotonicity constraint.
if Shape == 1 % Directly solve the problem in this case
    P = speye(T);
    % The A matrix for constraints along the first regressor
    Diff = eye(T0-1,T0) - [zeros(T0-1,1) eye(T0-1,T0-1)];
    Diff = kron(eye(T0),Diff);
    % For the second regressor, need to compute the commutation matrix
    Indx = reshape(1:T,T0,T0)';
    Commut = eye(T);
    Commut = Commut(Indx(:),:);
    % Now put them together
    A = sparse([Diff;Diff*Commut]);
    lb = -Inf(size(A,1),1);
    ub = sparse(size(A,1),1);
% For Shape in {2,3}, the configurations are based on the CNLS problem in
% Kuosmanen (2008, Econometrics Journal).    
else
    % Construct the coordinates of the grid points with the constants.
    [G1,G2] = meshgrid(Grid);
    FullGrid = [ones(T,1) G1(:) G2(:)];
    % Matrices and vectors for the quadratic programming problem.
    AuxDim = (Zdim+1)*T;         % Number of parameters in the CNLS prolem
    H = num2cell(FullGrid,2);
    H = sparse(blkdiag(H{:}));   % Design matrix for the CNLS problem
    A = zeros(T*(T-1),AuxDim);   % Full set of concavity constraints
    for t = 1:T
        TempGrid = FullGrid(t,:);
        jt = ((t-1)*(Zdim+1)+1):(t*(Zdim+1));
        Indx = ([1:t-1,t+1:T]);
        for s = Indx
            i = (t-1)*(T-1)+s;
            js = ((s-1)*(Zdim+1)+1):(s*(Zdim+1));
            if s < t
                A(i,js) = -TempGrid;
                A(i,jt) = TempGrid;
            else
                A(i-1,js) = -TempGrid;
                A(i-1,jt) = TempGrid;
            end
        end
    end
    P = sparse(H'*H);           % P for the quadratic term as in OSQP
    A = sparse(A);              % A is sparse
    if Shape == 3               % Stack the constraints
        A = [A; sparse(kron(eye(T),[zeros(Zdim,1) -eye(Zdim)]))];
    end
    lb = -Inf(size(A,1),1);     % Bounds of constraints for the solver
    ub = sparse(size(A,1),1);
end

% Specifying parameters for the solver
rho = 1/T;         % Based on Mazumder et al (2019, JASA)
Display = 0;       % Turn off standard messages from the solver
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Implement the Test %%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct the knot sequences (based on the equiquantile spacing interior 
% knots), and the tensor product B-splines.
Zknots = cell(Zdim,1);       % Where to store the knots
for d = 1:Zdim
    Zknots{d} = quantile(Z(:,d),0:1/(Zkn+1):1);
    Zknots{d} = augknt(Zknots{d},Ord);        % The augmented knots
    Sieve0 = spcol(Zknots{d}, Ord, Grid);     % Size: T*kn
    if d == 1
        Sieve = Sieve0;
    else
        Sieve = kron(Sieve, Sieve0);          % Final size: (T*kn)^Zdim
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct series functions via tensor product of B-splines
Hn = zeros(N,kn^Zdim);
for k = 1:Zdim
    [ZTemp, index] = sort(Z(:,k));         % Sorting required by spcol
    HnTemp = spcol(Zknots{k}, Ord, ZTemp); % N*kn
    HnTemp = sortrows([index HnTemp]);     % Arrange rows across obs
    Hn0 = HnTemp(:,2:end);                 % Size: N*kn
    if k == 1
        Hn(:,1:kn) = Hn0;
    else
        PreHn = Hn(:,1:kn^(k-1));
        Hn(:,1:kn^k) = reshape(reshape(PreHn,N,1,[]).*Hn0,N,[]);
    end
end

% Second, obtain the series LS estimators and residuals
Hatbeta = Hn\Y;            % LS estimator for coefficient
Hattheta = Sieve*Hatbeta;  % T*1: estimator for regression function
HatU = Y - Hn*Hatbeta;     % Residuals

% Third, obtain the constrained estimator through projection
if Shape == 1
    q = sparse(-Hattheta);      % The q vector in the OSQP problem
else
    q = sparse(-H'*Hattheta);   % The q vector in the OSQP problem
end
Prob = osqp;                    % Initialize the solver
settings = Prob.default_settings();
settings.rho = rho;
settings.verbose = Display;
Prob.setup(P,q,A,lb,ub,settings);
Results = Prob.solve();         % This will return a structure
Projtheta = Results.x;          % The projection as a minimizer
if Shape ~= 1
    Projtheta = min(FullGrid*reshape(Projtheta,Zdim+1,T),[],2); % Size: T*1
end

% Finally, compute the test statistic via numerical integration
% based on the trapezoidal integration
Diff = reshape(Hattheta - Projtheta,T0,T0); % Convert to do trapz
Stat = (sqrt(N)/Xin)*trapz(GridStep,trapz(GridStep,Diff.^2,2))^0.5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bootstrap Gn
BootW = randn(N,Boot);                        % Bootstrap weights
BGn = (sqrt(N)/Xin)*Sieve*(Hn\(BootW.*HatU)); % Score bootstrap: T*Boot
MultBGn = reshape(BGn,T0,T0,Boot);            % Reshape along each dim

% Next, obtain the data driven tuning parameter
BGnNorm = trapz(GridStep,trapz(GridStep,MultBGn.^2,2),1).^0.5;
Hattau = quantile(BGnNorm,1-gamma);           % Quantile
Hatkappa = sqrt(N)/(log(N)*Xin*Hattau);       % $\hat\kappa_n$

% Now, construct the bootstrap statistics
BootSum = BGn+Hatkappa*Projtheta;      % Bootstrap + drift: T*Boot
if Shape == 1
    BootProj = zeros(T,Boot);          % Store the projections
else
    BootTempProj = zeros(AuxDim,Boot); % Store the estimates for projections
end
% The boostrap loops
parfor i = 1:Boot
    if Shape == 1
        bootq = sparse(-BootSum(:,i));      % What to be projected
    else
        bootq = sparse(-H'*BootSum(:,i));   % What to be projected
    end
    Prob = osqp;                        % Initialize the solver
    settings = Prob.default_settings();
    settings.rho = rho;
    settings.verbose = Display;
    Prob.setup(P,bootq,A,lb,ub,settings);
    Results = Prob.solve();             % Just solve it!
    if Shape == 1
        BootProj(:,i) = Results.x;      % Projection as a minimizer
    else
        BootTempProj(:,i) = Results.x;  % Projection as a minimizer
    end
end
% Compute the projection by Kuosmanen (2008, Econometrics Journal)
if Shape ~= 1
    BootTempProj = reshape(BootTempProj,Zdim+1,T,Boot);
    BootPreProj = arrayfun(@(i) FullGrid*squeeze(BootTempProj(:,:,i)),...
        1:Boot,'Uni',0);
    BootPreProj = cat(3,BootPreProj{:});
    BootProj= squeeze(min(BootPreProj,[],2)); % T*Boot
end

% Compute the ultimate bootstrap statistic
BootDiff = reshape(BootSum-BootProj,T0,T0,Boot);
BStats = trapz(GridStep,trapz(GridStep,BootDiff.^2,2),1).^0.5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Record p-Value %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pvalue = mean(BStats >= Stat);

disp(['The p-value is ' num2str(pvalue)])


% Close parallel computing 
delete (gcp);