%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% Written by Zheng Fang -- March 2020 %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements Chetverikov's test for monotonicity of regression
% function (2019, ET), based on the code generously shared by the author.
% Here, we consider the univariate case.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                         % The significance level
N = 500;                              % The number of observations
Rep = 1000;                           % Number of Monte Carlo simulations
Boot = 200;                           % The number of bootstrap repetitions
bn = 0.2*(log(N))^0.5/N^0.25;         % Bandwidth for Rice's estimator
BootIndx = 'SD';                      % Choose the bootstrap method
gamma = 0.01;                         % Tuning parameter for set estimation

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
H0 = [0, 0.1, 0.5;                          % The null configurations
      0, 0.5, 2;
      0, 0.5, 1];
b = linspace(0.2,2,10);                     % Here go the alternatives
a = zeros(1,length(b));
c = 5*ones(1,length(b));
H1 = [a;b;c];                              
DGP = [H0, H1];                             % The full configurations
D = size(DGP,2);
a = DGP(1,:);
b = DGP(2,:);
c = DGP(3,:);
clear DGP H0 H1

% Next, some preallocations
Rej = zeros(D,Rep);                   % Record rejections 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parfor r = 1:Rep                               % Loop over MC replications
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Groudwork %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    % Data input for the current loop
    Z = -1+2*normcdf(randn(N,1));             % Regressor 
    U = randn(N,1);                           % Error
    
    % Allocate a temporary array to record rejections in the current loop
    TempRej = zeros(D,1);
    
    for j = 1:D                              % Loop over designs
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Generate the outcome for the current DGP 
        Y = a(j)*Z - b(j)*normpdf(c(j)*Z) + U; %#ok<PFBNS>        
        % First, obtain the b(s)-statistics based on the basic set of
        % weighting functions recommended in Chetverikov (2019,ET,p.736).
        hmax = range(Z)/2; 
        hmin = 0.4*hmax*(log(N)/N)^(1/3);
        Hn = exp(log(hmax):log(0.5):log(hmin))';% Let its length be h 
        LocalZ = Z*ones(1,N) - ones(N,1)*Z';    % N*N: Grids at actual obs
        RepLZ = repmat(LocalZ,[length(Hn),1]);  % Nh*N: Stack A to match Hn
        RepH = kron(Hn,ones(N,1));     % Match each entry of Hn to each A
        KerW = RepLZ./RepH;            % Nh*N: (Z_j-Z_i)/h_k
        KerW = 0.75*max(1-KerW.^2,0);  % Use Epanechnikov kernel
        W = KerW.*(KerW*sign(LocalZ)); % Nh*N
        Stats = W*Y;                   % Nh*1: b(s) is linear in Y's
        
        % Next, compute the local Rice variance estimator described in
        % Section 2.4 and Comment 4.1 in Chetverikov (2019,ET).
        LocalIndx = (abs(LocalZ) <= bn);  % The i-th row are local indices for sigma_i
        LocalIndx = LocalIndx(:,1:(N-1)); % j+1\in J(i) excludes the N-th column
        DiffY = Y(2:N) - Y(1:(N-1));
        VarError = (LocalIndx*DiffY.^2)./(2*sum(LocalIndx,2)); % N*1: Equation (13)
        
        % Now, studentize b(s) to obtain the test statistic
        Std = W.^2*VarError;              % Nh*1: V(s)
        Stats = Stats./sqrt(Std);         % Nh*1: Studentization
        Stat = max(Stats);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, compute the objectives of the bootstrap suprema       
        BootError = randn(N,Boot).*VarError.^0.5; % N*Boot: N(0,1) multipliers
        BootStats = W*BootError./sqrt(Std); % Nh*Boot: Bootstrapped b(s)

        % First, the critical value based on the plug-in method; see p.739.
        if strcmpi(BootIndx, 'PI')              
            CV_PI = quantile(max(BootStats),1-alpha);
            TempRej(j) = (Stat > CV_PI);
        % Next, the one-step critical value; see p.740.
        elseif strcmpi(BootIndx, 'OS')
            AuxCV_PI = quantile(max(BootStats),1-gamma);  
            OSindx = (Stats > -2*AuxCV_PI);
            OSBootStats = reshape(BootStats(repmat(OSindx,[1,Boot])),[],Boot);
            CV_OS = quantile(max(OSBootStats),1-alpha);
            TempRej(j) = (Stat > CV_OS);
        % Lastly, the step-down critical value; see p.741.
        else
            AuxCV_PI = quantile(max(BootStats),1-gamma); 
            OSindx = (Stats > -2*AuxCV_PI);
            OSBootStats = reshape(BootStats(repmat(OSindx,[1,Boot])),[],Boot);
            SDBootStat = max(OSBootStats);
            J1 = sum(OSindx);
            while true
                AuxCV_SD = quantile(SDBootStat,1-gamma);  
                SDindx = (Stats > -AuxCV_PI-AuxCV_SD); 
                SDBootStats = reshape(BootStats(repmat(SDindx,[1,Boot])),[],Boot);
                SDBootStat = max(SDBootStats);
                J2 = sum(SDindx);
                if J2 == J1
                    break
                else
                    J1 = J2;
                end
            end
            TempRej(j) = (Stat > quantile(SDBootStat,1-alpha));            
        end            
    end
    Rej(:,r) = TempRej;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%% Save the Results and Clean up %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fid = fopen(['OutputMC1_N' num2str(N) 'C' BootIndx 'Rep' num2str(Rep)...
    'Boot' num2str(Boot) '.txt'],'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
delete (gcp);