%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% Written by Zheng Fang -- March 2020 %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements LSW's test for monotonicity of the regression
% function in NP models, based on their L2 test statistic.

% The null hypotheses is that first derivative of the regression function
% is nonnegative. We conduct local polynomial estimation of order p = 2 as
% recommended by Fan & Gijbels (the 1996 book, p.59).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                         % The significance level
N = 500;                              % Number of observations
Rep = 1000;                           % Number of Monte Carlo simulations
Boot = 200;                           % Number of bootstrap repetitions
Shape = 3;                            % 1 for monotonicity, 2 convexity, 3 both
eta = 0.001;                          % The infinitesimal uniformity factor 
GridStep = 0.05;                      % Step size for discretization
Grid = (-0.9:GridStep:0.9)';          % The vector of grid points
T = length(Grid);                     % Number of grid points
Smooth = 'Un';                        % Smoothing: Un-under and Op-optimal
Student = 1;                          % Studentization or not
DataDriven = 0;                       % Indicator for data-driven bandwidth
Consth = 2;                           % Constant for the bandwidth

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
H0 = [0, 0.1, 0.5;                          % The null configurations
      0, 0.5, 2;
      0, 0.5, 1];
b = linspace(0.2,2,10);                     % Here go the alternatives
a = zeros(1,length(b));
c = 5*ones(1,length(b));
H1 = [a;b;c];                              
DGP = [H0, H1];                             % The full configurations
D = size(DGP,2);
a = DGP(1,:);
b = DGP(2,:);
c = DGP(3,:);

clear DGP H0 H1

% Select the order of the local polynomial regression
if Shape == 1            
    Ord = 2;
elseif Shape == 2
    Ord = 3;
elseif Shape == 3
    Ord = 3;
else
    error('Try again; Shape must be one of 1, 2 and 3.')
end

% Next, some preallocations
Rej = zeros(D,Rep);                         % Record rejections 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parfor r = 1:Rep                            % Loop over MC replications
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Groudwork %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    % Data input for the current loop
    Z = -1+2*normcdf(randn(N,1));         % Regressor
    U = randn(N,1);                       % Error

    for j = 1:D                           % Loop over designs
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%% Some Groundwork %%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Generate the outcome for the current DGP
        Y = a(j)*Z - b(j)*normpdf(c(j)*Z) + U; %#ok<PFBNS>
        % Obtain the bandwidth by the rule of thumb bandwidth prescribed by 
        % (4.3) in Fan & Gijbels (1996), for the data driven choice.
        if DataDriven == 1
            p = polyfit(Z,Y,Ord+3);      % First, a polynomial fit
            VarError = mean(zscore(Y-polyval(p,Z)).^2);
            pp = polyder(polyder(p));
            Dfit = polyval(pp,Z);        % The 2nd derivative of the fit
            h0 = 2.275*(VarError*1.8/mean((Dfit.^2).*(abs(Z)<=0.9)))^(1/(2*Ord+3))...
                *N^(-1/(2*Ord+3));
        else
            h0 = N^(-1/(2*Ord+3));
        end
        % Obtain the final bandwidth given h0
        if strcmpi(Smooth,'Un')
            h = Consth*std(Z)*h0*N^(1/(2*Ord+3))*N^(-1/(2*(Ord-1)+3));
        else
            h = Consth*std(Z)*h0;
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%% Compute the Test Statistic %%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, obtain the derivative estimator (with Epanechnikov kernel)
        % Some preallocations
        if Shape ~= 3
            Stats = zeros(T,1); % 1st/2nd derivative  
            if Student  == 1
                Std = zeros (T,1);
            end
        else                    % 1st and 2nd derivatives  
            Stats = zeros(T,2);
            if Student  == 1
                Std = zeros (T,2);
            end
        end
        % Obtain the estimators evaluated at grid points 
        for t = 1:T
            W = diag(1.5*max(1-(2*(Z-Grid(t))/h).^2,0));  %#ok<PFBNS>
            WY = sqrt(W)*Y;
            LPZ = (Z-Grid(t)).^(0:Ord);  % Local powers of ZZ
            WZ = sqrt(W)*LPZ;
            Coeff = (WZ\WY)';
            if Shape ~= 3
                Stats(t) = factorial(Ord-1)*Coeff(Ord);
            else
                Stats(t,:) = [factorial(Ord-2) factorial(Ord-1)].*Coeff(Ord-1:Ord);
            end
            % If selected, obtain the standard errors for the regression
            % estimator based on a local polynomial fit of order p+2. This 
            % is based on p.115 in Fan & Gijbels (1996).
            if Student == 1
                Auxh = Consth*std(Z)*N^(-1/(2*(Ord+2)+3));
                AuxW = diag(1.5*max(1-(2*(Z-Grid(t))/Auxh).^2,0));
                AuxWY = sqrt(AuxW)*Y;
                AuxLPZ = [LPZ,(Z-Grid(t)).^(Ord+1:Ord+2)];                
                AuxWZ = sqrt(AuxW)*AuxLPZ;
                AuxErrorSq = diag(AuxW)'*(Y - AuxWZ*(AuxWZ\AuxWY)).^2;
                VarError = AuxErrorSq/trace(AuxW-AuxW*AuxLPZ*...
                    ((AuxLPZ'*AuxW*AuxLPZ)\AuxLPZ')*AuxW);
                V = (LPZ'*W*LPZ)\LPZ'*W*W*LPZ/(LPZ'*W*LPZ)*VarError;
                if Shape ~= 3
                    Std(t) = factorial(Ord-1)*sqrt(V(Ord,Ord));
                else
                    Std(t,:) = [factorial(Ord-2) factorial(Ord-1)].*...
                                 [sqrt(V(Ord-1,Ord-1)),sqrt(V(Ord,Ord))];
                end
            end
        end
        
        % Now, compute the L2 statistic via numerical integration. First,
        % compute integrand (by incorporating the contact set).
        if Student == 0
            if Shape ~= 3
            rn = sqrt(N*h^(1+2*(Ord-1)));
            Stat = trapz(GridStep, max(-rn*Stats,0).^2);
            else
            rn = [sqrt(N*h^(1+2*(Ord-2))),sqrt(N*h^(1+2*(Ord-1)))];
            Stat = trapz(GridStep, max([-Stats.*rn zeros(T,1)],[],2).^2);                
            end
        else
            Stat = trapz(GridStep, max([-Stats./Std zeros(T,1)],[],2).^2);
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%% Obtain Critical Value via Bootstrap %%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, obtain the bootstrap estimators: the same process as for 
        % the test statistics but based on bootstrap samples
        if Shape ~= 3
            BootStats = zeros(T,1,Boot);  % Store bootstrap estimates
        else
            BootStats = zeros(T,2,Boot);  % Store bootstrap estimates
        end
        for i = 1:Boot
            % Efron's empirical bootstap as in LSW
            BootIndx = ceil(N*rand(N,1));
            BZ = Z(BootIndx);
            BY = Y(BootIndx);
            for t = 1:T
                BW = diag(1.5*max(1-(2*(BZ-Grid(t))/h).^2,0));
                BWY = sqrt(BW)*BY;
                BWZ = sqrt(BW)*((BZ-Grid(t)).^(0:Ord));
                BCoeff = (BWZ\BWY)';
                if Shape ~= 3
                    BootStats(t,:,i) = factorial(Ord-1)*BCoeff(Ord);
                else
                    BootStats(t,:,i) = [factorial(Ord-2) factorial(Ord-1)].*...
                                        BCoeff(Ord-1:Ord);
                end
            end
        end
        
        % Second, obtain the tuning parameter for estimating contact set
        if Student == 0
            BootG = -(BootStats-Stats).*rn;
        else
            BootG = -(BootStats-Stats)./Std;
        end
        Cn = quantile(max(max(BootG,[],[1,2]),sqrt(log(N))),1-0.1/log(N));
        Cn = 0.4*log(log(N))*Cn;          % See the display (5.1) in LSW
        % Third, compute the bootstrap statistic
        if Shape ~= 3
            if Student == 0
                ConSet = (abs(rn*Stats)<= Cn);
            else
                ConSet = (abs(Stats./Std)<= Cn);
            end
            BootG = squeeze(max(BootG,0).^2).*ConSet;
            BootStat = trapz(GridStep,BootG);
        else
            if Student == 0
                EP1 = rn(1)*Stats(:,1);
                EP2 = rn(2)*Stats(:,2);
            else
                EP1 = Stats(:,1)./Std(:,1);
                EP2 = Stats(:,2)./Std(:,2);
            end
            ConSet1 = (abs(EP1)<= Cn & EP2<-Cn);         % Size: T*1
            ConSet2 = (abs(EP2)<= Cn & EP1<-Cn);         % Size: T*1
            ConSet12 = (abs(EP1)<= Cn & abs(EP2)<= Cn);  % Size: T*1
            % Third, compute the bootstrap statistic
            BootG1 = max(squeeze(BootG(:,1,:)),0).^2.*ConSet1;
            BootG2 = max(squeeze(BootG(:,2,:)),0).^2.*ConSet2;
            BootG12 = squeeze(max(max(BootG,[],2),0)).^2.*ConSet12;
            BootStat = trapz(GridStep,BootG1+BootG2+BootG12);
        end
        
        % Now, we are ready to compute the critical value
        CV = max(quantile(BootStat,1-alpha),sqrt(h)*eta+mean(BootStat));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%% Record Rejection %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        Rej(j,r) = (Stat > CV);             
        
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%% Save the Results and Clean up %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fid = fopen(['OutputMC1_LSW_N' num2str(N) 'Shape' num2str(Shape) 'Rep'...
    num2str(Rep) 'Boot' num2str(Boot) 'GridStep' num2str(GridStep) ...
    'DataDriven' num2str(DataDriven) 'Constant' num2str(Consth) 'Std'...
    num2str(Student) Smooth 'Smoothing.txt'],'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
delete (gcp);