%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Written by Zheng Fang and Juwon Seo -- April 2020 %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements our shape tests of the regression function in NP 
% models with multiple regressors, based on B-splines. The "Shape" 
% indicator allows one to choose different shapes to test (monotonicity, 
% convexity, or both). We solve the quadratic problems by the solver OSQP 
% (https://osqp.org/).

% A (very) brief introduction to B-splines may be found in Appendix E.1 in
% Chen & Christensen (2018, QE). One may also want to consult de Boor's "A
% Practical Guide to Splines" (2001), in particular Theorem 44 on p.97 and
% Figure 50 on p.100, to better understand the commmands augknt and spcol.

% The MATLAB Engine API for Python allows one to call MATLAB as a 
% computational engine from Python; see instructutions at
% https://www.mathworks.com/products/matlab/matlab-and-python.html.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end  
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                  % The significance level
N = 1000;                      % Number of observations
Rep = 3000;                    % Number of MC replications 
Boot = 200;                    % Number of bootstrap repetitions
Shape = 1;                     % 1 for monotonicity, 2 concavity, 3 both
Zdim = 2;                      % Dimension of the regressors
Zkn = 1;                       % Number of interior knots for B-splines
Ord = 3;                       % Order of B-splines 
kn = Zkn+Ord;                  % Sieve dimension for Z
Xin = sqrt(kn^Zdim);           % Norm bound of the tensor product B-splines
if Shape == 1
    GridStep = 0.05;           % Step size for discretization
else
    GridStep = 0.1;            % Step size for discretization
end
Grid = (0.1:GridStep:0.9)';    % The vector of grid points
T0 = length(Grid);             % Number of grid points along each dimension
T = T0^Zdim;                   % Total number of grid points
Bandwidth = 'L3';              % Select the tuning parameter
if strcmpi(Bandwidth,'C1')
    gamma = 0.1;                     
elseif strcmpi(Bandwidth,'C2')
    gamma = 0.05;
elseif strcmpi(Bandwidth,'C3')
    gamma = 0.01;
elseif strcmpi(Bandwidth,'L1')
    gamma = 0.1/log(N);
elseif strcmpi(Bandwidth,'L2')
    gamma = 0.05/log(N);
elseif strcmpi(Bandwidth,'L3')
    gamma = 0.01/log(N);
elseif strcmpi(Bandwidth,'P1')
    gamma = 1/N^0.5;
elseif strcmpi(Bandwidth,'P2')
    gamma = 1/N^0.75;
elseif strcmpi(Bandwidth,'P3')
    gamma = 1/N;
else
    error('Try a different Bandwidth')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
H0 = [0, 0.2, 0.5;                               % The null configurations
      0, 1,   0;
      0, 0,   0.5];
if Shape == 1
    a = -linspace(0.05,0.5,10);                  % The alternatives
elseif Shape == 2
    a = -linspace(0.2,2,10);                     % The alternatives
elseif Shape == 3
    a = -linspace(0.05,0.5,10);                  % The alternatives
else
    error('Shape must be in {1, 2, 3}.')
end
b = zeros(1,length(a));                         
c = a;
H1 = [a;b;c];                              
DGP = [H0, H1];  
D = size(DGP,2);
a = DGP(1,:);
b = DGP(2,:);
c = DGP(3,:);
clear H0 H1 DGP

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Next, configurations for projection depending on the shape selected.
% Start with the monotonicity constraint.
if Shape == 1 % Directly solve the problem in this case
    P = speye(T);
    % The A matrix for constraints along the first regressor
    Diff = eye(T0-1,T0) - [zeros(T0-1,1) eye(T0-1,T0-1)];
    Diff = kron(eye(T0),Diff);
    % For the second regressor, need to compute the commutation matrix
    Indx = reshape(1:T,T0,T0)';
    Commut = eye(T);
    Commut = Commut(Indx(:),:);
    % Now put them together
    A = sparse([Diff;Diff*Commut]);
    lb = -Inf(size(A,1),1);
    ub = sparse(size(A,1),1);
% For Shape in {2,3}, the configurations are based on the CNLS problem in
% Kuosmanen (2008, Econometrics Journal).    
else
    % Construct the coordinates of the grid points with the constants.
    [G1,G2] = meshgrid(Grid);
    FullGrid = [ones(T,1) G1(:) G2(:)];
    % Matrices and vectors for the quadratic programming problem.
    AuxDim = (Zdim+1)*T;         % Number of parameters in the CNLS prolem
    H = num2cell(FullGrid,2);
    H = sparse(blkdiag(H{:}));   % Design matrix for the CNLS problem
    A = zeros(T*(T-1),AuxDim);   % Full set of concavity constraints
    for t = 1:T
        TempGrid = FullGrid(t,:);
        jt = ((t-1)*(Zdim+1)+1):(t*(Zdim+1));
        Indx = ([1:t-1,t+1:T]);
        for s = Indx
            i = (t-1)*(T-1)+s;
            js = ((s-1)*(Zdim+1)+1):(s*(Zdim+1));
            if s < t
                A(i,js) = -TempGrid;
                A(i,jt) = TempGrid;
            else
                A(i-1,js) = -TempGrid;
                A(i-1,jt) = TempGrid;
            end
        end
    end
    P = sparse(H'*H);           % P for the quadratic term as in OSQP
    A = sparse(A);              % A is sparse
    if Shape == 3               % Stack the constraints
        A = [A; sparse(kron(eye(T),[zeros(Zdim,1) -eye(Zdim)]))];
    end
    lb = -Inf(size(A,1),1);     % Bounds of constraints for the solver
    ub = sparse(size(A,1),1);
end

% Specifying parameters for the solver
rho = 1/T;         % Based on Mazumder et al (2019, JASA)
Display = 0;       % Turn off standard messages from the solver

% Finally, some preallocations
Rej = zeros(D,Rep);                      % Record rejections

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parfor r = 1:Rep             % Loop over MC replications (columns first)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Groudwork %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    % Data for the current loop
    Z = normcdf(randn(N,Zdim));  % Regressor
    U = randn(N,1);              % Error
    % Construct the knot sequences (based on the equiquantile spacing 
    % interior knots), and the tensor product B-splines.
    Zknots = cell(Zdim,1);       % Where to store the knots
    for d = 1:Zdim
        Zknots{d} = quantile(Z(:),0:1/(Zkn+1):1);
        [Zknots{d}(1), Zknots{d}(end)] = deal(0,1);
        Zknots{d} = augknt(Zknots{d},Ord);        % The augmented knots
        Sieve0 = spcol(Zknots{d}, Ord, Grid);     % Size: T*kn
        if d == 1
            Sieve = Sieve0;
        else
            Sieve = kron(Sieve, Sieve0);          % Final size: (T*kn)^Zdim
        end
    end
    
    for j = 1:D               % Loop over designs (then come the rows)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, construct the matrix of tensor product B-splines evaluated
        % at the data points, in a way compatible with Grid
        % Generate the outcome based on the design
        if Shape ~= 2
            LogTerm = c(j)*log(1+sum(Z,2));         %#ok<PFBNS>
        else
            LogTerm = c(j)*log(1+5*sum(Z,2));             
        end
        if b(j) == 0                                %#ok<PFBNS>
            Y = a(j)*prod(sqrt(Z),2) + LogTerm + U; %#ok<PFBNS>
        else
            Y = a(j)*(sum(0.5*Z.^(b(j)),2)).^(1/b(j)) + LogTerm + U;
        end
        % Construct series functions via tensor product of B-splines
        Hn = zeros(N,kn^Zdim);
        for k = 1:Zdim
            [ZTemp, index] = sort(Z(:,k));      % Sorting required by spcol
            HnTemp = spcol(Zknots{k}, Ord, ZTemp); % N*kn
            HnTemp = sortrows([index HnTemp]);  % Arrange rows across obs
            Hn0 = HnTemp(:,2:end);              % Size: N*kn
            if k == 1
                Hn(:,1:kn) = Hn0;
            else
                PreHn = Hn(:,1:kn^(k-1));
                Hn(:,1:kn^k) = reshape(reshape(PreHn,N,1,[]).*Hn0,N,[]);
            end
        end     

        % Second, obtain the series LS estimators and residuals
        Hatbeta = Hn\Y;            % LS estimator for coefficient
        Hattheta = Sieve*Hatbeta;  % T*1: estimator for regression function
        HatU = Y - Hn*Hatbeta;     % Residuals
        
        % Third, obtain the constrained estimator through projection
        if Shape == 1
            q = sparse(-Hattheta);      % The q vector in the OSQP problem 
        else
            q = sparse(-H'*Hattheta);   % The q vector in the OSQP problem 
        end
        Prob = osqp;                    % Initialize the solver
        settings = Prob.default_settings();
        settings.rho = rho;
        settings.verbose = Display;
        Prob.setup(P,q,A,lb,ub,settings);
        Results = Prob.solve();         % This will return a structure
        Projtheta = Results.x;          % The projection as a minimizer
        if Shape ~= 1
            Projtheta = min(FullGrid*reshape(Projtheta,Zdim+1,T),[],2); % Size: T*1
        end
        
        % Finally, compute the test statistic via numerical integration
        % based on the trapezoidal integration
        Diff = reshape(Hattheta - Projtheta,T0,T0); % Convert to do trapz
        Stat = (sqrt(N)/Xin)*trapz(GridStep,trapz(GridStep,Diff.^2,2))^0.5;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Bootstrap Gn         
        BootW = randn(N,Boot);                        % Bootstrap weights 
        BGn = (sqrt(N)/Xin)*Sieve*(Hn\(BootW.*HatU)); % Score bootstrap: T*Boot
        MultBGn = reshape(BGn,T0,T0,Boot);            % Reshape along each dim
        
        % Next, obtain the data driven tuning parameter
        BGnNorm = trapz(GridStep,trapz(GridStep,MultBGn.^2,2),1).^0.5;
        Hattau = quantile(BGnNorm,1-gamma);           % Quantile
        Hatkappa = sqrt(N)/(log(N)*Xin*Hattau);       % $\hat\kappa_n$  
        
        % Now, construct the bootstrap statistics
        BootSum = BGn+Hatkappa*Projtheta;      % Bootstrap + drift: T*Boot
        if Shape == 1
            BootProj = zeros(T,Boot);          % Store the projections
        else
            BootTempProj = zeros(AuxDim,Boot); % Store the estimates for projections
        end
        % The boostrap loops  
        for i = 1:Boot
            if Shape == 1
                bootq = sparse(-BootSum(:,i));      % What to be projected 
            else
                bootq = sparse(-H'*BootSum(:,i));   % What to be projected
            end
                Prob = osqp;                        % Initialize the solver
                settings = Prob.default_settings();
                settings.rho = rho;
                settings.verbose = Display;
                Prob.setup(P,bootq,A,lb,ub,settings);
                Results = Prob.solve();             % Just solve it!
                if Shape == 1
                    BootProj(:,i) = Results.x;      % Projection as a minimizer
                else
                    BootTempProj(:,i) = Results.x;  % Projection as a minimizer
                end
        end
        % Compute the projection by Kuosmanen (2008, Econometrics Journal)
        if Shape ~= 1
            BootTempProj = reshape(BootTempProj,Zdim+1,T,Boot);
            BootPreProj = arrayfun(@(i) FullGrid*squeeze(BootTempProj(:,:,i)),...
                1:Boot,'Uni',0);
            BootPreProj = cat(3,BootPreProj{:});
            BootProj= squeeze(min(BootPreProj,[],2)); % T*Boot
        end
        
        % Compute the ultimate bootstrap statistic
        BootDiff = reshape(BootSum-BootProj,T0,T0,Boot);
        BStats = trapz(GridStep,trapz(GridStep,BootDiff.^2,2),1).^0.5;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%% Record Rejection %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        Rej(j,r) = (Stat > quantile(BStats,1-alpha));     
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Save Results %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
output = ['OutputMC2_N' num2str(N) 'Shape' num2str(Shape) 'Rep' ...
   num2str(Rep) 'Boot' num2str(Boot) 'GridStep' num2str(GridStep) 'Order'... 
   num2str(Ord) 'Knots' num2str(Zkn) Bandwidth '.txt'];
fid = fopen(output,'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
% delete (gcp);