%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% Written by Zheng Fang -- May 2020 %%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements Chetverikov's test for monotonicity of regression
% function (2019, ET), based on the code generously shared by the author.
% Here, we consider the bivariate case.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                         % The significance level
N = 1000;                             % Number of observations
Rep = 3000;                           % Number of Monte Carlo simulations
Boot = 200;                           % Number of bootstrap repetitions
Zdim = 2;
bn = 0.2*(log(N))^0.5/N^0.25;         % Bandwidth for Rice's estimator
BootIndx = 'SD';                      % Choose the bootstrap method
gamma = 0.01;                         % Tuning parameter for set estimation
GridStep = 0.2;                       % Spacing of location points

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
H0 = [0, 0.2, 0.5;                             % The null configurations
      0, 1,   0;
      0, 0,   0.5];
a = -linspace(0.05,0.5,10);                    % The alternatives
b = zeros(1,length(a));                         
c = a;
H1 = [a;b;c];                              
DGP = [H0, H1];  
D = size(DGP,2);
a = DGP(1,:);
b = DGP(2,:);
c = DGP(3,:);
clear H0 H1 DGP

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = zeros(D,Rep);                            % Record rejections 

parfor r = 1:Rep                               % Loop over MC replications
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% Preparations: Location Points, Bandwidths and Weights %%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Data inputs for the current loop
    Z = normcdf(randn(N,Zdim));
    U = randn(N,1);
    % Create location points and bandwidths
    LZ1 = quantile(Z(:,1),0.1:GridStep:0.9);  % Quantiles as grid for Z1
    LZ2 = quantile(Z(:,2),0.1:GridStep:0.9);  % Quantiles as grid for Z2
    [LZ1,LZ2] = meshgrid(LZ1,LZ2);
    PairZ = [LZ1(:) LZ2(:)];               % Coordinates of location points
    NumLZ = size(PairZ,1);
    Indxh = cell(Zdim,1);
    for d = 1:Zdim
        hmax = range(Z(:,d))/2;                     
        hmin = 0.4*hmax*(log(N)/N)^(1/3);           
        Indxh{d} = exp(log(hmax):log(0.5):log(hmin))'; 
    end
    [Indxh1,Indxh2] = meshgrid(Indxh{1},Indxh{2});
    Pairh = [Indxh1(:) Indxh2(:)];   % What we want for bandwidths
    Numh = size(Pairh,1);
    % Compute the weights based on location points and widths. These are
    % weights such that the top b(s) on p.27 of the working paper equals
    % sum_i Y_i W_i(s) with W_i(s) = sum_j sgn(X_j-X_i)Q_{ij}(s)/2 almost
    % surely if Z_i's are continuous
    W = zeros(N,Numh,NumLZ);
    for i = 1:NumLZ
        TempW = (Z - PairZ(i,:))./reshape(Pairh',1,2,Numh);
        TempW = prod(0.75*max(1-TempW.^2,0),2); % Epanechnikov kernel
        TempW = TempW.*permute(TempW,[2,1,3]);
        W(:,:,i) = squeeze(sum(sign(Z(:,1)-Z(:,1)').*TempW/2,1)); 
    end
    
    % Generate the outcomes based on the designs
    TempRej = zeros(D,1);    
    for d = 1:D                              % Loop over designs
        % To begin with, generate the outcome in the currnet loop        
        LogTerm = c(d)*log(1+sum(Z,2));             %#ok<PFBNS>
        if b(d) == 0                                %#ok<PFBNS>
            Y = a(d)*prod(sqrt(Z),2) + LogTerm + U; %#ok<PFBNS>
        else
            Y = a(d)*(sum(0.5*Z.^b(d),2)).^(1/b(d)) + LogTerm + U;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, compute the local Rice variance estimator described in
        % Section 2.4 and Comment 4.1 in Chetverikov (2019,ET).
        LocalIndx = (squeeze(vecnorm(Z-reshape(Z',1,[],N),2,2))' <= bn);
        LocalIndx = LocalIndx(:,1:(N-1)); % Exclude the N-th column since j+1\in J(i)
        DiffY = Y(2:N) - Y(1:(N-1));
        VarError = (LocalIndx*DiffY.^2)./(2*sum(LocalIndx,2)); % N*1: Equation (13)
        
        % Next, compute the objectives in test and bootstrap statistics
        Stats = zeros(NumLZ*Numh,1);  
        BootError = randn(N,Boot).*VarError.^0.5;
        BootStats = zeros(NumLZ*Numh,Boot);
        for i = 1:NumLZ
            Std = (VarError'*W(:,:,i).^2).^0.5;       % Size: 1*Numh
            Stats(((i-1)*Numh+1:i*Numh)) = (Y'*W(:,:,i))./Std;
            BootStats(((i-1)*Numh+1:i*Numh),:) = (W(:,:,i)'*BootError)./Std';
        end
        Stat = max(Stats);   % Stats (a vector indexed by s) vs Stat 

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, the critical value based on the plug-in method; see p.739.
        if strcmpi(BootIndx, 'PI')              
            CV_PI = quantile(max(BootStats),1-alpha);
            TempRej(d) = (Stat > CV_PI);
        % Next, the one-step critical value; see p.740.
        elseif strcmpi(BootIndx, 'OS')
            AuxCV_PI = quantile(max(BootStats),1-gamma);  
            OSindx = (Stats > -2*AuxCV_PI);
            OSBootStats = reshape(BootStats(repmat(OSindx,[1,Boot])),[],Boot);
            CV_OS = quantile(max(OSBootStats),1-alpha);
            TempRej(d) = (Stat > CV_OS);
        % Lastly, the step-down critical value; see p.741.
        else
            AuxCV_PI = quantile(max(BootStats),1-gamma);  
            OSindx = (Stats > -2*AuxCV_PI);
            OSBootStats = reshape(BootStats(repmat(OSindx,[1,Boot])),[],Boot);
            SDBootStat = max(OSBootStats);
            J1 = sum(OSindx);
            while true
                AuxCV_SD = quantile(SDBootStat,1-gamma);  
                SDindx = (Stats > -AuxCV_PI-AuxCV_SD); 
                SDBootStats = reshape(BootStats(repmat(SDindx,[1,Boot])),[],Boot);
                SDBootStat = max(SDBootStats); % Update SDBootStat
                J2 = sum(SDindx);
                if J2 == J1
                    break
                else
                    J1 = J2;
                end
            end
            TempRej(d) = (Stat > quantile(SDBootStat,1-alpha));            
        end            
    end
    Rej(:,r) = TempRej;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%% Save the Results and Clean up %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%           
fid = fopen(['OutputMC2_N' num2str(N) 'C' BootIndx 'Rep' num2str(Rep)...
    'Boot' num2str(Boot) '.txt'],'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
delete (gcp);