%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% Written by Zheng Fang -- March 2020 %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements LSW's shape (monotonicity/concavity) tests for the 
% regression function in bivariate NP models, based on their L2 statistics.

% The null hypothese is that partial derivatives of the regression function
% are nonnegative. Thus, we conduct local polynomial estimation with p = 2
% or 3 as recommended by Fan & Gijbels (the 1996 book, p.59).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                         % The significance level
N = 1000;                             % Number of observations
Rep = 1000;                           % Number of Monte Carlo simulations
Boot = 200;                           % Number of bootstrap repetitions
Shape = 1;                            % 1 for monotonicity, 2 convexity 
Zdim = 2;                             % Number of regressors
eta = 0.001;                          % The infinitesimal uniformity factor 
if Shape == 1
    GridStep = 0.05;                  % Step size for discretization
else
    GridStep = 0.1;                   % Step size for discretization
end
Grid = (0.1:GridStep:0.9)';           % The vector of grid points
T0 = length(Grid);                    % Number of grid points along each dimension
T = T0^Zdim;                          % Total number of grid points
NumTau = 500;             % Number of MC draws to evaluate integral wtr tau
Smooth = 'Un';            % Smoothing: Un-under and Op-optimal
Student = 1;              % Studentization
DataDriven = 0;           % Indicator for data driven bandwidth
Consth = 2;               % Constant for the bandwidth

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
H0 = [0, 0.2, 0.5;                               % The null configurations
      0, 1,   0;
      0, 0,   0.5];
if Shape == 1
    a = -linspace(0.05,0.5,10);                  % The alternatives
elseif Shape == 2
    a = -linspace(0.2,2,10);                     % The alternatives
else
    error('Shape must be in {1, 2}.')
end
b = zeros(1,length(a));                         
c = a;
H1 = [a;b;c];                              
DGP = [H0, H1];  
D = size(DGP,2);
a = DGP(1,:);
b = DGP(2,:);
c = DGP(3,:);
clear H0 H1 DGP;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Select the order of the local polynomial regression
if Shape == 1            
    Ord = 2;
    Res = eye(Ord+4);
    Res = Res(2:3,:);
    fac = diag([1,1]);
    A = fac*Res;
    clear Res fac;
elseif Shape == 2
    Ord = 3;
    Res = eye(Ord+7);
    Res = Res(4:6,:);
    fac = diag([2,2,1]);
    A = fac*Res;
    clear Res fac;
else
    error('Try again; Shape must be one of 1 and 2.')
end

% Create grid points to evaluate functions
[G1,G2] = meshgrid(Grid);
FullGrid = [G1(:) G2(:)];
clear G1 G2;

% Next, some preallocations
Rej = zeros(D,Rep);                             % Record rejections 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parfor r = 1:Rep             % Loop over MC replications (columns first)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Groundwork %%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%      
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    % Data input for the current loop
    Z = normcdf(randn(N,Zdim));            % Regressor
    U = randn(N,1);                        % Error
    % Obtain the bandwidth by the rule of thumb, as follows.   
    if strcmpi(Smooth,'Un')
        h = Consth*std(Z,0,'all')*N^(-1/(Zdim+2*(Ord-1)+2));
    else
        h = Consth*std(Z,0,'all')*N^(-1/(Zdim+2*Ord+2));
    end
    % The bandwidth for variance estimation (based on local polynomial
    % estimation with order Ord+2).
    if Student == 1
        Auxh = Consth*std(Z,0,'all')*N^(-1/(Zdim+2*(Ord+2)+2));
    end
    
    for j = 1:D                              % Loop over designs      
        % First, construct the matrix of tensor product B-splines evaluated
        % at the data points, in a way compatible with Grid
        % Generate the outcome
        if Shape == 1
            LogTerm = c(j)*log(1+sum(Z,2));              %#ok<*PFBNS>
        else
            LogTerm = c(j)*log(1+5*sum(Z,2));             
        end
        if b(j) == 0                                 
            Y = a(j)*prod(sqrt(Z),2) + LogTerm + U;  
        else
            Y = a(j)*(sum(0.5*Z.^(b(j)),2)).^(1/b(j)) + LogTerm + U;
        end
        % MC draws for evaluating integration wtr tau
        MCTau = unifrnd(0,2*pi,NumTau,1);
        GridTau = [cos(MCTau) sin(MCTau)];
        GridTau = [GridTau.^2,2*prod(GridTau,2)];

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%% Compute the Test Statistic %%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, obtain the derivative estimators (with the Epanechnikov
        % kernel). Begin with some preallocations.
        if Shape == 1
            Stats = zeros(T,Zdim);      % First derivatives
            if Student == 1
                Std = zeros(T,Zdim);
            end
        else 
            Stats = zeros(T,NumTau);    % Second derivatives
            if Student == 1
                Std = zeros(T,NumTau);
            end         
        end
        % Now loop over evaluation points for actual estimation.
        for t = 1:T
            Temp = Z-FullGrid(t,:);            
            W = sparse(diag(prod(0.75*max(1-(Temp/h).^2,0),2))); 
            WY = sqrt(W)*Y; 
            if Shape == 1
                LZ = [ones(N,1) Temp Temp.^2 prod(Temp,2)]; % Size: N*6
            else
                LZ = [ones(N,1) Temp Temp.^2 prod(Temp,2) Temp.^3 ... % N*10
                    (Temp(:,1).^2).*Temp(:,2) Temp(:,1).*(Temp(:,2).^2)];
            end
            WZ = sqrt(W)*LZ;
            Coeff = WZ\WY;
            if Shape == 1
                Stats(t,:) = (A*Coeff)';
            else 
                Stats(t,:) = (GridTau*(A*Coeff))';           
            end
            % Compute standard errors (Fan & Gijbels, 1996, p.115).
            if Student == 1
                AuxW = sparse(diag(prod(0.75*max(1-(Temp/Auxh).^2,0),2))); 
                AuxWY = sqrt(AuxW)*Y;
                if Shape == 1
                    AuxLZ = [LZ Temp.^3 (Temp(:,1).^2).*Temp(:,2)...
                        Temp(:,1).*(Temp(:,2).^2) Temp.^4 (Temp(:,1).^3).*Temp(:,2)...
                        Temp(:,1).^2.*(Temp(:,2).^2) Temp(:,1).*(Temp(:,2).^3)];
                else
                    AuxLZ = [LZ Temp.^3 (Temp(:,1).^2).*Temp(:,2)...
                        Temp(:,1).*(Temp(:,2).^2) Temp.^4 (Temp(:,1).^3).*Temp(:,2)...
                        Temp(:,1).^2.*(Temp(:,2).^2) Temp(:,1).*(Temp(:,2).^3)...
                        Temp.^5 (Temp(:,1).^4).*Temp(:,2) (Temp(:,1).^3).*(Temp(:,2).^2)...
                        Temp(:,1).^2.*(Temp(:,2).^3) Temp(:,1).*(Temp(:,2).^4)];
                end
                AuxWZ = sqrt(AuxW)*AuxLZ;
                AuxErrorSq = diag(AuxW)'*(Y - AuxWZ*(AuxWZ\AuxWY)).^2;
                VarError = AuxErrorSq/trace(AuxW-AuxW*AuxLZ*...
                    ((AuxLZ'*AuxW*AuxLZ)\AuxLZ')*AuxW);
                V = (LZ'*W*LZ)\LZ'*W*W*LZ/(LZ'*W*LZ)*VarError;
                if Shape == 1
                    Std(t,:) = (diag(sqrt(A*V*A')))';
                else 
                    Std(t,:) = (diag(sqrt(GridTau*A*V*A'*GridTau')))';
                end                
            end
        end
        % Now, compute the L2 statistic via numerical integration. First,
        % compute integrands (by incorporating the contact sets).
        if Student == 0
            if Shape == 1
                rn = sqrt(N*h^4);
                TempStats = reshape(max([-rn*Stats zeros(T,1)],[],2).^2,T0,T0);
            else 
                rn = sqrt(N*h^6);
                TempStats = reshape(mean(max(rn*Stats,0).^2,2),T0,T0);         
            end
        elseif Student == 1
            if Shape == 1
                TempStats = reshape(max([-Stats./Std zeros(T,1)],[],2).^2,T0,T0);
            else 
                TempStats = reshape(mean(max(Stats./Std,0).^2,2),T0,T0);                             
            end
        else
            error('Student must be either 0 or 1.')
        end
        Stat = trapz(GridStep, trapz(GridStep,TempStats,2));
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%% Obtain Critical Value via Bootstrap %%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, obtain the bootstrap estimators : the same process as for 
        % the test statistics but based on bootstrap samples
        if Shape == 1
            BootStats = zeros(T,Zdim,Boot);
        else 
            BootStats = zeros(T,NumTau,Boot);
        end
        % Loop over bootstrap repetitions
        for i = 1:Boot
            % Efron's empirical bootstap as in LSW
            BootIndx = ceil(N*rand(N,1));
            BZ = Z(BootIndx,:);
            BY = Y(BootIndx);
            for t = 1:T
                Temp = BZ-FullGrid(t,:);
                BW = sparse(diag(prod(0.75*max(1-(Temp/h).^2,0),2)));
                BWY = sqrt(BW)*BY;
                if Shape == 1
                    BLZ = [ones(N,1) Temp Temp.^2 prod(Temp,2)];
                else
                    BLZ = [ones(N,1) Temp Temp.^2 prod(Temp,2) Temp.^3 ...
                        (Temp(:,1).^2).*Temp(:,2) Temp(:,1).*(Temp(:,2).^2)];
                end
                BWZ = sqrt(BW)*BLZ;
                BCoeff = BWZ\BWY;
                if Shape == 1
                    BootStats(t,:,i) = (A*BCoeff)';
                else 
                    BootStats(t,:,i) = (GridTau*(A*BCoeff))';
                end
            end
        end
        
        % Second, obtain the tuning parameter for estimating contact set
        if Student == 0
            if Shape == 1
                BootG = -rn*(BootStats-Stats);   
            else 
                BootG = rn*(BootStats-Stats); 
            end
        else
            if Shape == 1
                BootG = -(BootStats-Stats)./Std;   
            else 
                BootG = (BootStats-Stats)./Std;                
            end
        end
        Cn = quantile(max(max(BootG,[],[1,2]),sqrt(log(N))),1-0.1/log(N));
        Cn = 0.4*log(log(N))*Cn;          % See the display (5.1) in LSW
        if Shape == 1
            if Student == 0
                EP1 = rn*Stats(:,1);
                EP2 = rn*Stats(:,2);
            else
                EP1 = Stats(:,1)./Std(:,1);
                EP2 = Stats(:,2)./Std(:,2);
            end
            ConSet1 = (abs(EP1)<=Cn & EP2<-Cn);       
            ConSet2 = (abs(EP2)<=Cn & EP1<-Cn);       
            ConSet12 = (abs(EP1)<=Cn & abs(EP2)<=Cn);       
            % Third, compute the bootstrap statistic
            BootInt1 = max(squeeze(BootG(:,1,:)),0).^2.*ConSet1;
            BootInt2 = max(squeeze(BootG(:,2,:)),0).^2.*ConSet2;
            BootInt12 = squeeze(max(max(BootG,[],2),0)).^2.*ConSet12;
            BootInt = reshape(BootInt1+BootInt2+BootInt12,T0,T0,Boot);
        else 
            if Student == 0
                Conset = (abs(rn*Stats)<=Cn);
            else
                Conset = (abs(Stats./Std)<=Cn);
            end
            BootInt = reshape(mean(max(BootG,0).^2.*Conset,2),T0,T0,Boot);
        end
        BootStat = trapz(GridStep,trapz(GridStep,BootInt,2),1);
        
        % Now, we are ready to compute the critical value
        CV = max(quantile(BootStat,1-alpha),sqrt(h)*eta+mean(BootStat));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%% Record Rejection %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        Rej(j,r) = (Stat > CV);             
        
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%% Save the Results and Clean up %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fid = fopen(['OutputMC2_LSW_N' num2str(N) 'Shape' num2str(Shape) ...
     'Rep' num2str(Rep) 'Boot' num2str(Boot) 'NumTau' num2str(NumTau) ...
     'DataDriven' num2str(DataDriven) 'GridStep' num2str(GridStep) ...
     'Constant' num2str(Consth) 'Std' num2str(Student) Smooth 'Smoothing.txt'],'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
delete (gcp);