%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Written by Zheng Fang and Juwon Seo -- June 2020 %%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This module implements our test of the Slutsky restriction for a system
% of partially linear models with multiple regressors, based on B-splines.

% A (very) brief introduction to B-splines may be found in Appendix E.1 in
% Chen & Christensen (2018, QE). One may also want to consult de Boor's "A
% Practical Guide to Splines" (2001), in particular Theorem 44 on p.97 and
% Figure 50 on p.100, to better understand the commmands augknt and spcol.

% The MATLAB Engine API for Python allows one to call MATLAB as a 
% computational engine from Python; see instructutions at
% https://www.mathworks.com/products/matlab/matlab-and-python.html.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Set the Stage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Turn off warnings
clear
if isempty(gcp('nocreate'))
    warning('off','all')
else
    pctRunOnAll warning('off', 'all')
end
% Choose a random stream for reproducibility
stream = RandStream.create('Threefry','Seed',310);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% Set up Parameters for the Simulation %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 0.05;                  % The significance level
N = 5000;                      % Number of observations
Rep = 3000;                    % Number of MC replications 
Boot = 200;                    % Number of bootstrap repetitions
GDim = 2;                      % Dimension of the goods: 2 or 3
ZDim = 1;                      % Dimension of controls
Pkn0 = 0;                      % Number of interior knots for prices
POrd = 3;                      % Order of B-splines for prices
Pkn = Pkn0+POrd;               % Number of basis for univariate price
Ykn0 = 0;                      % Number of interior knots for income
YOrd = 3;                      % Order of B-splines for income
Ykn = Ykn0+YOrd;               % Number of basis for univariate income
kn = Pkn^GDim*Ykn;             % Sieve dimension
Xin = sqrt(kn);                % Norm bound of the tensor product B-splines
Bandwidth = 'L3';              % Select the tuning parameter
if strcmpi(Bandwidth,'C1')
    gamma = 0.1;                   
elseif strcmpi(Bandwidth,'C2')
    gamma = 0.05;
elseif strcmpi(Bandwidth,'C3')
    gamma = 0.01;
elseif strcmpi(Bandwidth,'L1')
    gamma = 0.1/log(N);
elseif strcmpi(Bandwidth,'L2')
    gamma = 0.05/log(N);
elseif strcmpi(Bandwidth,'L3')
    gamma = 0.01/log(N);
elseif strcmpi(Bandwidth,'P1')
    gamma = 1/N^0.5;
elseif strcmpi(Bandwidth,'P2')
    gamma = 1/N^0.75;
elseif strcmpi(Bandwidth,'P3')
    gamma = 1/N;
else
    error('Try a different Bandwidth')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Simulation Designs %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
PMin = 1;                      % Left endpoint for price
PMax = 2;                      % Right endpoint
YMin = 0;                      % Left endpoint for income
YMax = 1;                      % Right endpoint
delta = [1 1 1 0.1:0.1:1];     % Localizaiton parameter
D = length(delta);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Preparations %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create grid points for evaluation of the norms
PGridStep = 0.05;                        % Step size for discretization
PGrid = (PMin+0.1:PGridStep:PMax-0.1)';  % Grid points for price
YGridStep = 0.05;                        % Step size for discretization
YGrid = (YMin+0.1:YGridStep:YMax-0.1)';  % Grid points for income
% Obtain the coordinates for the grid points
if GDim == 1
    [G1,G2] = meshgrid(PGrid,YGrid);
    FullGrid = [G1(:) G2(:)];
    clear G1 G2;
elseif GDim == 2
    [G1,G2,G3] = ndgrid(PGrid,PGrid,YGrid);
    FullGrid = [G1(:) G2(:) G3(:)];
    clear G1 G2 G3;
elseif GDim == 3
    [G1,G2,G3,G4] = ndgrid(PGrid,PGrid,PGrid,YGrid);
    FullGrid = [G1(:) G2(:) G3(:) G4(:)];
    clear G1 G2 G3 G4;
else
    error('GDim = 1, 2 or 3 only; modify the code before proceeding.')
end
M = size(FullGrid,1);                % Number of grid points
MultiM = [length(PGrid)*ones(1,GDim) length(YGrid)]; % Numbers of grid points along each dimension
clear PGrid YGrid;

% Some vectors and matrices that faciliate computations later.
Multikn0 = [Pkn0*ones(1,GDim) Ykn0]; % Vector of numbers of interior knots
MultiOrd = [POrd*ones(1,GDim) YOrd]; % Vector of orders of B-splines 
Multikn = [Pkn*ones(1,GDim) Ykn];    % Vector of univariate sieve dimension
DerOrd = eye(GDim+1);                % Multi-index for partial derivatives

% Specify coefficients for the B-form of the tensor product B-splines that
% we use to approximate demand functions. Check out de Boor for B-forms.
Coefs = zeros([kn Multikn]);        % As required for the command spmak
if GDim == 1
    for k = 1:Ykn
        for d = 1:Pkn
            indx = sub2ind(Multikn,d,k);
            Coefs(indx,d,k) = 1;
        end
    end
elseif GDim == 2
    for k = 1:Ykn
        for d2 = 1:Pkn
            for d1 = 1:Pkn
                indx = sub2ind(Multikn,d1,d2,k);
                Coefs(indx,d1,d2,k) = 1;
            end
        end
    end
elseif  GDim == 3
    for k = 1:Ykn
        for d3 = 1:Pkn
            for d2 = 1:Pkn
                for d1 = 1:Pkn
                    indx = sub2ind(Multikn,d1,d2,d3,k);
                    Coefs(indx,d1,d2,d3,k) = 1;
                end
            end
        end
    end
end

% Finally, some preallocations
Rej = zeros(D,Rep);                      % Record rejections

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% Now, Start the Loops %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parfor r = 1:Rep             % Loop over MC replications (columns first)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%% Some Groudwork %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Choose substreams for reproducibility
    set(stream,'Substream',r);
    RandStream.setGlobalStream(stream);
    % Data input for the current loop
    P = PMin + (PMax - PMin)*normcdf(randn(N,GDim)); % Prices
    Y = YMin + (YMax - YMin)*normcdf(randn(N,1));    % Income
    PY = [P Y];                                     
    Z = normcdf(randn(N,ZDim));                      % Controls
    U = randn(N,GDim);                               % Errors
    
    % Create the augmented knot sequences for all
    Knots = cell(GDim+1,1);     % Where to store the augmented knots
    for d = 1:GDim+1
        knots0 = quantile(PY(:,d),0:1/(Multikn0(d)+1):1); %#ok<PFBNS>
        if d ~= GDim+1
            [knots0(1), knots0(end)] = deal(PMin,PMax);
        else
            [knots0(1), knots0(end)] = deal(YMin,YMax);
        end
        Knots{d} = augknt(knots0,MultiOrd(d));           %#ok<PFBNS>  
    end
    
    % Construct the tensor product B-splines (based on the equiquantile
    % spacing interior knots) evaluated at the obs
    Bform = spmak(Knots,Coefs); % The kn*1 vector of basis in B-form
    Hn = fnval(Bform,PY')';     % Evaluate the basis at obs
    
    for j = 1:D                 % Loop over designs (then come the rows)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%% Compute the Statistic %%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, generate the outcomes
        if j == 1                       % DGP1: The boundary case
            Q = ones(N,GDim)/GDim + repmat(sum(Z,2),1,GDim) + U;
        elseif j == 2                   % DGP2: The C-D case
            Q = (Y./P)/GDim + repmat(sum(Z,2),1,GDim) + U;
        elseif j == 3                   % DGP3: The CES case with rho = 0.5
            Q = (Y./sum(1./P,2))./(P.^2) + repmat(sum(Z,2),1,GDim) + U;
        % The alternatives:
        else
            Q = exp((P-1.5).*[delta(j) -delta(j)]) + repmat(sum(Z,2),1,GDim) + U; %#ok<PFBNS>
        end

        % Second, obtain the series LS estimators
        HatCoefs = [Hn Z]\Q;            % Coefficients
        HatU = Q - [Hn Z]*HatCoefs;     % Residuals
        HatCoefs = reshape(HatCoefs(1:kn,:)',[GDim Multikn]); % For spmak
        
        % Third, obtain demand fn's and the derivatives via their B-forms
        HatBform = spmak(Knots,HatCoefs); % This is $\hat g_n$ in B-form
        HatDemand = reshape(fnval(HatBform,FullGrid'),1,GDim,[]); 
        DerDemand = cell(GDim+1,1);
        for d = 1:GDim+1
            DerDemand{d} = fnval(fnder(HatBform,DerOrd(d,:)),FullGrid'); %#ok<PFBNS>
        end
        % Convert the demand and its derivatives in desired sizes
        DpHatDemand = reshape(cat(1,DerDemand{1:end-1}),GDim,GDim,[]);
        DyHatDemand = reshape(DerDemand{end},GDim,1,[]);
        
        % In turn, compute the Slutsky matrix
        Stats = DpHatDemand + DyHatDemand.*HatDemand; % Size: GDim*GDim*M
        
        % Finally, compute the test statistic via numerical integration
        rn = sqrt(N)/Xin;
        Temp = arrayfun(@(m) eig((Stats(:,:,m)+Stats(:,:,m)')/2),1:M,'Uni',0);
        Temp = sum(max(cell2mat(Temp),0).^2,1); % The integrand
        Temp = reshape(Temp,MultiM);
        if GDim == 1
            Stat = rn*trapz(PGridStep,trapz(YGridStep,Temp,2))^0.5;
        elseif GDim == 2
            Stat = rn*trapz(PGridStep,trapz(PGridStep,trapz(...
                YGridStep,Temp,3),2))^0.5;
        else % Gdim == 3
            Stat = rn*trapz(PGridStep,trapz(PGridStep,trapz(...
                PGridStep,trapz(YGridStep,Temp,4),3),2))^0.5;
        end
      
        % Free up memory as we go
        DerDemand = []; DpHatDemand = []; 
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%% Obtain Critical Value through Bootstrap %%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % First, initialize arrays where we store bootstrap statistics
        BootStats = zeros(GDim,GDim,M,Boot); % The bootstrap matrix mapping
        BootNorm = zeros(Boot,1);         % Norms of BootStats (for kappan)
        
        % Loop over bootstrap repetitions
        for i = 1:Boot
            % Estimation based on bootstrap samples
            BootW = randn(N,1);      % Standard normal weights
            StarCoefs = [Hn Z]\(BootW.*HatU);
            StarCoefs = rn*reshape(StarCoefs(1:kn,:)',[GDim Multikn]);
            
            % Sepcify the B-form for estimating W and its derivatives
            StarBform = spmak(Knots,StarCoefs);
            HatWn = reshape(fnval(StarBform,FullGrid'),1,GDim,[]);
            DerDemand = cell(GDim+1,1);
            for d = 1:GDim+1
                DerDemand{d} = fnval(fnder(StarBform,DerOrd(d,:)),FullGrid'); 
            end
            DpHatWn = reshape(cat(1,DerDemand{1:end-1}),GDim,GDim,[]);
            DyHatWn = reshape(DerDemand{end},GDim,1,[]);
            
            % Compute the bootstrap matrix mapping $\hat Gn$
            Temp = DpHatWn + DyHatDemand.*HatWn + DyHatWn.*HatDemand;
            BootStats(:,:,:,i) = Temp;
            
            % Compute the norm of Temp (as a matrix-valued mapping)
            Temp = squeeze(sum(Temp.^2,[1 2])); % Frobenius norm at each point
            Temp = reshape(Temp,MultiM);
            if GDim == 1
                BootNorm(i) = trapz(PGridStep,trapz(YGridStep,Temp,2))^0.5;
            elseif GDim == 2
                BootNorm(i) = trapz(PGridStep,trapz(PGridStep,...
                                         trapz(YGridStep,Temp,3),2))^0.5;
            else % Gdim == 3
                BootNorm(i) = trapz(PGridStep,trapz(PGridStep,...
                      trapz(PGridStep,trapz(YGridStep,Temp,4),3),2))^0.5;
            end
        end
        
        % Free up memomry as we go
        HatWn = []; DyHatWn = []; DpHatWn = []; Temp = []; HatDemand = [];
        DyHatDemand = []; BootW = [];
        
        % Next, obtain the data driven tuning parameter
        Hattau = quantile(BootNorm,1-gamma); 
        Hatkappa = rn/(log(N)*Hattau);
        
        % Finally, compute norms of the projected bootstrap statistics
        BootSum = BootStats+Hatkappa*Stats;             % T*Boot
        BootProjNorm = zeros(Boot,1);
        for i = 1:Boot
            Temp = arrayfun(@(m) eig((BootSum(:,:,m,i) + ...
                                        BootSum(:,:,m,i)')/2),1:M,'Uni',0);
            Temp = sum(max(cell2mat(Temp),0).^2,1);
            Temp = reshape(Temp,MultiM);
            if GDim == 1
                BootProjNorm(i) = trapz(PGridStep,trapz(YGridStep,...
                    Temp,2))^0.5;
            elseif GDim == 2
                BootProjNorm(i) = trapz(PGridStep,trapz(PGridStep,...
                    trapz(YGridStep,Temp,3),2))^0.5;
            else % Gdim == 3
                BootProjNorm(i) = trapz(PGridStep,trapz(PGridStep,...
                    trapz(PGridStep,trapz(YGridStep,Temp,4),3),2))^0.5;
            end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%% Record Rejection %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        Rej(j,r) = (Stat > quantile(BootProjNorm,1-alpha));     
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Finally, Compute the Empirical Reject Rates %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Rej = mean(Rej,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Save Results %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
output = ['OutputMC3_N' num2str(N) 'Rep' num2str(Rep) 'Boot' num2str(Boot)...
  'GDim' num2str(GDim) 'POrder' num2str(POrd) 'Pknots' num2str(Pkn0) ... 
  'YOrder' num2str(YOrd) 'Yknots' num2str(Ykn0) Bandwidth '.txt'];
fid = fopen(output,'w');
fprintf(fid,'%.4f\n', Rej);
fclose(fid);

% Close parallel computing 
delete (gcp);